#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: transformer.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-03-31
'''

import copy
import math
from functools import wraps

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.base import TransformerMixin

from .statistics import probability, WOE
from .utils import split_points_to_bin, FILLNA, save_json, split_target

DEFAULT_NAME = 'default_name'
EMPTY_BIN = -1
ELSE_GROUP = 'else'


def df_exclude_cols(func):
    @wraps(func)
    def exclude_cols(self, X, y, **kwargs):
        exclude = kwargs.get('exclude', None)
        if exclude is not None and isinstance(X, pd.DataFrame):
            X = X.drop(columns=exclude)
            del kwargs['exclude']

        return func(self, X, y, **kwargs)

    return exclude_cols


def df_select_dtypes(func):
    @wraps(func)
    def select_dtypes(self, X, y, **kwargs):
        select_dtypes = kwargs.get('select_dtypes', None)
        if select_dtypes is not None and isinstance(X, pd.DataFrame):
            X = X.select_dtypes(include=select_dtypes)
            del kwargs['select_dtypes']

        return func(self, X, y, **kwargs)

    return select_dtypes


def _check_duplicated_keys(X):
    if isinstance(X, pd.DataFrame) and X.columns.has_duplicates:
        keys = X.columns[X.columns.duplicated()].values
        raise Exception("X has duplicate keys `{keys}`".format(keys=str(keys)))

    return True


class FeatureBin(TransformerMixin):
    """
    分箱类
    """

    def __init__(self, n_jobs=-1):
        self.splits_dict = dict()
        self.n_jobs = n_jobs

    def __len__(self):
        return len(self.splits_dict.keys())

    def __contains__(self, key):
        return key in self.splits_dict

    def __getitem__(self, key):
        return self.splits_dict[key]

    def __setitem__(self, key, value):
        self.splits_dict[key] = value

    def __iter__(self):
        return iter(self.splits_dict)

    @df_exclude_cols
    @df_select_dtypes
    def fit(self, X, y, update = False, **kwargs):
        """
        分箱
        Args:
            X (DataFrame|array-like): 要分箱的X
            y (str|array-like): 目标变量
            min_sample_rate (number) : 每个箱子的最小占比，默认0.05
            n_bins (int): 需要分成几箱，默认10
            is_need_monotonic (bool) : 是否强制单调，默认True，强制单调
            is_empty_bin (bool) : 是否将空箱单独归为一个箱子，默认True，空箱单独归1箱
            min_threshold (number): 最小的卡方阀值
            exclude (str|array-like): 排除的特征，该特征将不参与分箱
            select_dtypes (str|numpy.dtypes): `'object'`, `'number'` 等. 只有选定的数据类型才会被分箱

        """

        # assert y.isin([0, 1]).all(), 'ERROR: :-) :-) 目标变量不是0/1值，请检查！！！'

        if not isinstance(X, pd.DataFrame):
            fea_name, splits = self._fit(X, y, **kwargs)
            self.splits_dict[fea_name] = splits
            return self

        if isinstance(y, str):
            # y = X.pop(y)
            X, y = split_target(X, y)

        _check_duplicated_keys(X)

        data = Parallel(n_jobs=self.n_jobs)(
            delayed(self._fit)(X[col], y, **kwargs) for col in X)  # 批量处理

        if update:
            self.splits_dict.update(dict(data))
        else:
            self.splits_dict = dict(data)

        return self

    def _fit(self, X, y, method='chi', is_empty_bin=True, **kwargs):  # method='dt'修改为method='chi'
        """
        分箱
        Args:
            X (DataFrame|array-like): 要分箱的X
            y (str|array-like): 目标变量
            min_sample_rate (number) : 每个箱子的最小占比
            n_bins (int): 需要分成几箱
            is_need_monotonic (bool) : 是否强制单调
            is_empty_bin (bool) : 是否将空箱单独归为一个箱子
            min_threshold (number): 最小的卡方阀值

        Returns:
            str : 分箱变量名
            array : 分割点

        """

        from .feature_binning import bin_method_run

        fea_name = DEFAULT_NAME
        if hasattr(X, 'name'):
            fea_name = X.name

        # 判断是否连续型变量，如果不是，将其原始值替换为0、1、2、3这样有序的连续性变量，序是按原始值所对应的woe值大小来给予的
        unique_X_val = None
        if not np.issubdtype(X.dtype, np.number):
            transer = WoeTransformer()
            # if X.dtype.type is np.object_:
            #     X = X.astype(np.str)
            empty_mask = pd.isna(X).any()
            if empty_mask:
                X = X.astype(np.str)
            woe = transer.fit_transform(X, y)
            # 获取变量的唯一值，及其所在该变量中的索引；unique_X_val=['A' 'B' 'C' 'D' 'E' 'F' 'G'] unique_X_index=[25  0  2  9  1  6 18]
            unique_X_val, unique_X_index = np.unique(X, return_index=True)
            # 通过原始值所在的索引将其原始对应的woe值取出。unique_woe=[-0.10178269 -0.44183275  0.22730944  0.15707894  0.50662326 -0.27946387 -0.10178269]
            unique_woe = woe[unique_X_index]
            # argsort函数是将woe值从小到大排序，然后将排序后的woe值所对应的原始woe值所在的索引输出；woe_argsort_index=[1 5 0 6 3 2 4]
            woe_argsort_index = np.argsort(unique_woe)
            # 变量唯一值按woe值从小到大的顺序调整变量唯一值的位置，变成有顺序的；unique_X_val=['B' 'F' 'A' 'G' 'D' 'C' 'E']
            unique_X_val = unique_X_val[woe_argsort_index]
            # 将原始X根据unique_X_val=['B' 'F' 'A' 'G' 'D' 'C' 'E']的有序顺序替换为0、1、2、3
            # unique_X_val=['B' 'F' 'G' 'nan' 'D' 'C' 'E']
            if empty_mask and is_empty_bin:
                unique_X_val = unique_X_val[np.isin(unique_X_val, 'nan', invert=True)]
            X = self._raw_category_x_to_bin(X, unique_X_val, is_empty_bin=is_empty_bin)

        splits = bin_method_run(X, y, method, is_empty_bin=is_empty_bin, **kwargs)

        # 如果不是连续型变量，X原始值被0、1、2、3替换了，自然出来的splits也是数值，需要将splits中的数值从unique_X_val=['B' 'F' 'A' 'G' 'D' 'C' 'E']进行还原
        splits = self._restore_category_splits(splits, unique_X_val)

        return fea_name, splits

    def transform(self, X, bins_dict={}, **kwargs):
        """
        原始数据根据分割点变换原始X
        Args:
            X (DataFrame|array-like): 需要转换的原始X
            bins_dict: 分箱字典, 形如: {'D157': [-999, 1.0, 2.0, 3.0, 5.0, inf]}
            **kwargs:

        Returns:

        """

        if not isinstance(bins_dict, dict):
            assert '请传入类似 {\'D157\': [-999, 1.0, 2.0, 3.0, 5.0, inf]}'

        if not bins_dict:
            bins_dict = self.splits_dict
        else:
            bins_dict = {k: np.array(v) for k, v in bins_dict.items()}

        if getattr(X, 'ndim', 1) == 1:

            if hasattr(X, 'name'):  # pd.Series
                if X.name in bins_dict:
                    fea_name, bins = self._transform(X, bins_dict.get(X.name), **kwargs)
                    return bins
                else:
                    return X

            if len(bins_dict) == 1:
                if DEFAULT_NAME in bins_dict:
                    fea_name, bins = self._transform(X, bins_dict.get(DEFAULT_NAME), **kwargs)
                    return bins
                else:
                    return X

        # X.reset_index(inplace=True)

        _check_duplicated_keys(X)

        ###并行处理
        data_with_bins = Parallel(n_jobs=self.n_jobs)(
            delayed(self._transform)(X[col], bins, **kwargs) for col, bins in bins_dict.items() if col in X)
        ###并行处理

        if isinstance(X, dict):
            return dict(data_with_bins)
        else:
            bin_df = pd.DataFrame(dict(data_with_bins), index=X.index)
            X_cols = list(X.columns)
            no_bin_cols = list(set(X_cols) - set(bin_df.columns))
            bin_df[no_bin_cols] = X[no_bin_cols]
            # X.set_index('index', inplace=True)

            # return bin_df.set_index('index')
            return bin_df[X_cols]

    def _transform(self, X, splits, labels=False):
        """
        原始数据根据分割点变换原始X
        Args:
            X (DataFrame|array-like): 需要转换的原始X
            splits (array) : 分割点
            labels (bool) : 转换后的X是否需要标签，默认False，不需要

        Returns:

        """

        fea_name = DEFAULT_NAME
        if hasattr(X, 'name'):
            fea_name = X.name

        if splits.ndim > 1 or not np.issubdtype(splits.dtype, np.number):
            empty_mask = pd.isna(X).any()
            if empty_mask:
                X = X.astype(np.str)
            bins = self._raw_category_x_to_bin(X, splits)

        else:
            bins = np.zeros(len(X), dtype=int)

            if len(splits):  # TODO 需要看下splits为什么会有空
                if np.isnan(splits[-1]):
                    mask = pd.isna(X)
                    bins[~mask] = split_points_to_bin(X[~mask], splits[:-1])
                    bins[mask] = len(splits)
                else:
                    bins = split_points_to_bin(X, splits)

        if labels:
            splits_format = self.splits_point_format(
                splits,
                index=True)  ## ['0.[-inf, 0.7470300495624542)' '1.[0.7470300495624542, inf)' '2.nan']  ['0.B,F,G' '1.D,C,E' '2.nan']
            mask = (bins == EMPTY_BIN)
            bins = splits_format[bins]
            bins[mask] = FILLNA

        return fea_name, bins
        # row = pd.Series({fea_name:bins})
        # row.index = X.index
        # return row

    def _raw_category_x_to_bin(self, X, unique_val, is_empty_bin=False):
        """
        原始变量进行转换
        Args:
            X (array-like): 需要转换的X
            unique_val (array-like): 分割点
            is_empty_bin (bool): 是否有空箱

        Returns:

        """
        if is_empty_bin:
            bins = np.full(len(X), np.nan)
        else:
            bins = np.full(len(X), EMPTY_BIN)
            # bins = np.full(len(X), len(unique_val) - 1)

        for i in range(len(unique_val)):
            val = unique_val[i]
            if isinstance(val, str) and val == ELSE_GROUP:
                bins[bins == EMPTY_BIN] = i
            else:
                bins[np.isin(X, val)] = i
        return bins

    def _restore_category_splits(self, splits, x_val):
        """
        将分割点复原回原始值
        Args:
            splits (array-like): 分割点
            x_val (array-like): 原始分割点

        Returns:
            array: 原回原始值分割点
        """
        if x_val is None:
            return splits

        empty_mask = np.isnan(splits).any()
        if empty_mask:
            splits = splits[~np.isnan(splits)]

        if isinstance(x_val, np.ndarray):
            x_val = x_val.tolist()

        restore_category_splits = []
        start = 0
        for i in splits:
            index = math.ceil(i)
            restore_category_splits.append(x_val[start:index])
            start = index

        restore_category_splits.append(x_val[start:])

        if empty_mask:
            restore_category_splits.append(['nan'])

        return np.array(restore_category_splits)

    def splits_point_format(self, splits, index=False, ellipsis=None, decimal=None):
        """
        将分割点格式化，形如：[0.[4 ~ 7), 1.[7 ~ 10)]
        Args:
            splits (array-like): 分割点
            index (bool): 是否需要下标，0.[4 ~ 7)中的0
            ellipsis:

        Returns:
            array: 格式化后的分割点
        """
        ## 数值型：splits=[0.45343156        nan]，类别型：splits=[list(['B', 'F', 'G']) list(['D', 'C', 'E']) list(['nan'])]
        l = list()

        if not np.issubdtype(splits.dtype, np.number):
            # for i in splits:
            #     l.append(','.join(i))
            for i in splits:
                if isinstance(i, str) and i == ELSE_GROUP:
                    l.append(i)
                else:
                    label = ','.join(i)
                    if ellipsis is not None:
                        label = label[:ellipsis] + '..' if len(label) > ellipsis else label
                    l.append(label)

        else:
            is_empty_split = len(splits) > 0 and np.isnan(splits[-1])  # TODO 需要看下splits为什么会有空
            if is_empty_split:
                splits = splits[:-1]

            splits_ = [-np.inf] + splits.tolist() + [np.inf]
            for i in range(len(splits_) - 1):
                l.append('['.format(i) + str(splits_[i]) + ' ~ ' + str(splits_[i + 1]) + ')')

            if is_empty_split:
                l.append('nan')

        if index:
            l = ["{}.{}".format(i, v) for i, v in enumerate(l)]

        return np.array(l)

    def manual_bin(self, manual_set_dict):
        """
        手动分箱
        Args:
            manual_set_dict (dict|array-like): map结构的分箱点，形如: {'D157': [1.0, 2.0, 3.0, 5.0]}

        Returns:

        """
        if not isinstance(manual_set_dict, dict):
            manual_set_dict = {
                DEFAULT_NAME: manual_set_dict,
            }

        assert isinstance(manual_set_dict, dict), '请传入类似 {\'D157\': [1.0, 2.0, 3.0, 5.0]}'

        for i in manual_set_dict:
            self.splits_dict[i] = np.array(manual_set_dict[i])
        return self

    def export(self, to_dataframe=False, to_json=None, to_csv=None, bin_format=True, index=True):
        """
        导出规则到dict或json或csv文件
        Args:
            to_dataframe (bool): 是否导出成pd.DataFrame形式
            to_json (str): 保存成json的路径
            to_csv (str): 保存成csv的路径
            bin_format (bool): 是否将分割点格式化
            index (bool): 分割点格式化是否需要下标

        Returns:
            dict: 分割点规则字典

        """

        splits = copy.deepcopy(self.splits_dict)
        if bin_format:
            # splits = {k: self.splits_point_format(v, index=False).tolist() for k, v in splits.items()}
            splits = {k: list(self.splits_point_format(v, index=index)) for k, v in splits.items()}
        else:
            splits = {k: v.tolist() for k, v in splits.items()}
        if to_json is not None:
            save_json(splits, to_json)

        if to_dataframe or to_csv is not None:

            row = []
            for var_name in splits:
                for bin in splits[var_name]:
                    row.append({
                        'feature': var_name,
                        'bins': bin
                    })

            splits = pd.DataFrame(row)

        if to_csv is not None:
            splits.to_csv(to_csv, index=False)

        return splits


class WoeTransformer(TransformerMixin):
    """
    woe转换类
    """

    def __init__(self, n_jobs=-1):
        self.fea_woe_dict = dict()
        self.n_jobs = n_jobs

    def __len__(self):
        return len(self.fea_woe_dict.keys())

    def __contains__(self, key):
        return key in self.fea_woe_dict

    def __getitem__(self, key):
        return self.fea_woe_dict[key]

    def __setitem__(self, key, value):
        self.fea_woe_dict[key] = value

    def __iter__(self):
        return iter(self.fea_woe_dict)

    @df_exclude_cols
    @df_select_dtypes
    def fit(self, X, y, update = False):
        """
        woe转换
        Args:
            X (DataFrame|array-like): 需要转换woe的X
            y (str|array-like): 目标变量
            exclude (str|array-like): 排除的变量，该变量将不参与woe计算
            select_dtypes (str|numpy.dtypes): `'object'`, `'number'` 等. 只有选定的数据类型才会被计算

        Returns:

        """
        if not isinstance(X, pd.DataFrame):
            fea_name, value_woe = self._fit_woe(X, y)
            self.fea_woe_dict[fea_name] = value_woe
            return self

        if isinstance(y, str):
            # y = X.pop(y)
            X, y = split_target(X, y)

        _check_duplicated_keys(X)

        data = Parallel(n_jobs=self.n_jobs)(
            delayed(self._fit_woe)(X[col], y) for col in X)  # 批量处理

        if update:
            self.fea_woe_dict.update(dict(data))
        else:
            self.fea_woe_dict = dict(data)

        return self

    def _fit_woe(self, X, y):
        """
        woe转换
        Args:
            X (DataFrame|array-like): 需要计算woe的X
            y (str|array-like): 目标变量

        Returns:
            str : 变量名
            array : 计算出来的woe值

        """
        fea_name = DEFAULT_NAME
        if hasattr(X, 'name'):
            fea_name = X.name
        X = np.copy(X)  # Series, array 转 np.array
        if X.dtype.type is np.object_:
            X = X.astype(np.str)
        unique_val = np.unique(X)

        # # TODO 如果X是连续性变量，且有空。计算出来的woe不正确
        # if X.dtype.type is np.object_:
        #     X = X.astype(np.str)
        #     unique_val = np.unique(X)
        # else:
        #     unique_val = [int(i) for i in np.unique(X)]

        value_woe = dict()
        for val in unique_val:
            y_prob, n_prob = probability(y, mask=(X == val))  #
            value_woe[val] = WOE(y_prob, n_prob)

        return fea_name, value_woe

    def transform(self, X, fea_woe_dict={}, **kwargs):
        """
        将原始值用woe值替换
        Args:
            X (DataFrame|array-like): 需要转换woe的X
            fea_woe_dict (dict): 变量和woe值的字典，形如：{'D157': {0: -0.46554351769099783, 1: -0.10263802400162944}}
            **kwargs:

        Returns:
            DataFrame: 转换woe后的X

        """

        if not isinstance(fea_woe_dict, dict):
            assert """请传入类似 {'D157': {'0.[-inf, 1.5)': -0.46554351769099783, '1.[1.5, 2.5)': -0.10263802400162944, '2.[2.5, 3.5)': 0.9591358533174893, '3.[3.5, 4.5)': 1.115806812932841, '4.[4.5, 7.5)': 1.1319717497861965, '5.[7.5, inf)': 2.369093204627806, '6.nan': -1.2516811662966312}}"""

        if not fea_woe_dict:
            fea_woe_dict = self.fea_woe_dict

        if getattr(X, 'ndim', 1) == 1:

            if hasattr(X, 'name'):  # pd.Series
                if X.name in fea_woe_dict:
                    fea_name, woe = self._transform(X, fea_woe_dict.get(X.name), fea_name=X.name)
                    return woe
                else:
                    return X

            if len(fea_woe_dict) == 1:
                if DEFAULT_NAME in fea_woe_dict:
                    fea_name, woe = self._transform(X, fea_woe_dict.get(DEFAULT_NAME))
                    return woe
                else:
                    return X

        # X.reset_index(inplace=True)

        _check_duplicated_keys(X)

        ###并行处理
        data_with_bins = Parallel(n_jobs=self.n_jobs)(
            delayed(self._transform)(X[col], woe_dict, fea_name=col, **kwargs) for col, woe_dict in fea_woe_dict.items()
            if col in X)
        ###并行处理

        if isinstance(X, dict):
            return dict(data_with_bins)
        else:
            bin_df = pd.DataFrame(dict(data_with_bins), index=X.index)
            X_cols = list(X.columns)
            no_bin_cols = list(set(X_cols) - set(bin_df.columns))
            bin_df[no_bin_cols] = X[no_bin_cols]
            # X.set_index('index', inplace=True)

            # return bin_df.set_index('index')
            return bin_df[X_cols]

    def _transform(self, X, woe_dict={}, fea_name=DEFAULT_NAME, other='min'):
        """

        Args:
            X (DataFrame|array-like): 需要转换woe的X
            woe_dict (dict): 变量和woe值的字典，形如：{0: -0.46554351769099783, 1: -0.10263802400162944}
            fea_name (str): 变量名
            other (str): 未来出现的新值给对应的woe值

        Returns:

        """

        try:
            woe = np.zeros(len(X))
        except:
            woe = np.zeros(np.array(X).size)

        if other == 'min':
            other = np.min(list(woe_dict.values()))
        elif other == 'max':
            other = np.max(list(woe_dict.values()))

        woe[np.isin(X, list(woe_dict.keys()), invert=True)] = other

        for k, v in woe_dict.items():
            woe[X == k] = v

        return fea_name, woe

    def load(self, manual_set_dict):
        """
        自定的woe值
        Args:
            manual_set_dict (dict): map结构的woe值，形如: {'D157': {0: -5.545177444479562, 1: 5.497168225293202}}

        Returns:

        """

        assert isinstance(manual_set_dict, dict), '请传入类似 {\'D157\': {0: -5.545177444479562, 1: 5.497168225293202}}'

        for i in manual_set_dict:
            self.fea_woe_dict[i] = manual_set_dict[i]
        return self

    def export(self, to_dataframe=False, to_json=None, to_csv=None, var_bin_woe={}):
        """
        导出规则到dict或json或csv文件
        Args:
            to_dataframe (bool): 是否导出成pd.DataFrame形式
            to_json (str): 保存成json的路径
            to_csv (str): 保存成csv的路径
            var_bin_woe (dict): {'D157': {0: -5.545177444479562, 1: 5.497168225293202}}

        Returns:
            dict: 分割点规则字典

        """

        if var_bin_woe:
            fea_bin_woe = var_bin_woe
        else:
            fea_bin_woe = copy.deepcopy(self.fea_woe_dict)
            #fea_bin_woe = {k: {int(i): j for i, j in v.items()} for k, v in fea_bin_woe.items()}
        if to_json is not None:
            save_json(fea_bin_woe, to_json)

        if to_dataframe or to_csv is not None:
            row = list()
            for var_name in fea_bin_woe:
                for bin, woe in fea_bin_woe[var_name].items():
                    row.append({
                        'feature': var_name,
                        'bins': bin,
                        'woe': woe
                    })

            fea_bin_woe = pd.DataFrame(row)

        if to_csv is not None:
            fea_bin_woe.to_csv(to_csv, index=False)

        return fea_bin_woe



#!/usr/bin/python

"""Command line tools for detecting csv data

Team: ESC

Examples:

    python detector.py -i xxx.csv -o report.csv

"""

import pandas as pd


def get_top_values(series, top=5, reverse=False):
    """Get top/bottom n values

    Args:
        series (Series): data series
        top (number): number of top/bottom n values
        reverse (bool): it will return bottom n values if True is given

    Returns:
        Series: Series of top/bottom n values and percentage. ['value:percent', None]
    """
    itype = 'top'
    counts = series.value_counts()
    counts = list(zip(counts.index, counts, counts.divide(series.size)))

    if reverse:
        counts.reverse()
        itype = 'bottom'

    template = "{0[0]}:{0[2]:.2%}"
    indexs = [itype + str(i + 1) for i in range(top)]
    values = [template.format(counts[i]) if i < len(counts) else None for i in range(top)]

    return pd.Series(values, index=indexs)


def get_describe(series, percentiles=[.25, .5, .75]):
    """Get describe of series

    Args:
        series (Series): data series
        percentiles: the percentiles to include in the output

    Returns:
        Series: the describe of data include mean, std, min, max and percentiles
    """
    d = series.describe(percentiles)
    return d.drop('count')


def count_blank(series, blanks=[None]):
    """Count number and percentage of blank values in series

    Args:
        series (Series): data series
        blanks (list): list of blank values

    Returns:
        number: number of blanks
        str: the percentage of blank values
    """
    # n = 0
    # counts = series.value_counts()
    # for blank in blanks:
    #     if blank in counts.keys():
    #         n += counts[blank]

    n = series.isnull().sum()

    return (n, "{0:.2%}".format(n / series.size), n / series.size)


def is_numeric(series):
    """Check if the series's type is numeric

    Args:
        series (Series): data series

    Returns:
        bool
    """
    return series.dtype.kind in 'ifc'


def detect(dataframe, dic_name=None):
    """ Detect data

    Args:
        dataframe (DataFrame): data that will be detected

    Returns:
        DataFrame: report of detecting
    """

    rows = []
    for name, series in dataframe.items():
        numeric_index = ['mean', 'std', 'min', '1%', '10%', '50%', '75%', '90%', '99%', 'max']
        discrete_index = ['top1', 'top2', 'top3', 'top4', 'top5', 'bottom5', 'bottom4', 'bottom3', 'bottom2', 'bottom1']

        details_index = [numeric_index[i] + '_or_' + discrete_index[i] for i in range(len(numeric_index))]
        details = []

        if is_numeric(series):
            desc = get_describe(
                series,
                percentiles=[.01, .1, .5, .75, .9, .99]
            )
            details = desc.tolist()
        else:
            top5 = get_top_values(series)
            bottom5 = get_top_values(series, reverse=True)
            details = top5.tolist() + bottom5[::-1].tolist()

        nblank, pblank, pblank_ = count_blank(series)

        ###add 2020/01/02 RyanZheng
        value_max_percent = get_max_percent(series)
        ###add 2020/01/02 RyanZheng

        row = pd.Series(
            index=['type', 'size', 'missing', 'missing_q', 'unique', 'value_max_percent'] + details_index,
            data=[series.dtype, series.size, pblank, pblank_, series.nunique(), value_max_percent] + details
        )

        row.name = name
        rows.append(row)

    # return pd.DataFrame(rows)

    ### add 2020/01/02 RyanZheng
    eda_df = pd.DataFrame(rows)
    if dic_name is not None and isinstance(dic_name, dict):
        # 增加一列中文名称列
        eda_df.insert(0, 'cn', eda_df.index.map(dic_name))
        # eda_df['cn'] = eda_df.index.map(dic_name)
    eda_df.index.name = 'var_name'
    eda_df = eda_df.reset_index()
    eda_df['type'] = eda_df['type'].astype(str)
    return eda_df
    ### add 2020/01/02 RyanZheng


###add 2020/01/02 RyanZheng
def get_max_percent(series):
    """
    获取变量中同一个值出现次数最多的该值的占比
    Args:
        series:

    Returns:

    """
    return max(series.value_counts(dropna=False) / len(series))
###add 2020/01/02 RyanZheng



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: plot.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-10-18
'''
import os
import re

import matplotlib.pyplot as plt
import numpy as np
import openpyxl
import pandas as pd
import seaborn as sns
import six
from matplotlib import gridspec
from openpyxl.drawing.image import Image
from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_curve, roc_auc_score, precision_recall_curve

from .statistics import calc_var_summary
from .utils import save_json


def ellipsis_fun(x, decimal=2, ellipsis=16):
    if re.search(r'\[(.*?) ~ (.*?)\)', str(x)):
        le = x.split('~')[0]
        ri = x.split('~')[1]
        left = re.match(r'\d+.\[(.*?) ', le)
        right = re.match(r' (.*?)\)', ri)
        if left:
            tmp = round(float(left.groups()[0]), decimal)
            l = f"{le.split('[')[0]}[{tmp} "
            le = l
        if right:
            tmp = round(float(right.groups()[0]), decimal)
            r = f" {tmp})"
            ri = r

        return f"{le} ~ {ri}"
    else:
        return str(x)[:ellipsis] + '..'


def plot_var_bin_summary(frame, cols, target='target', by='type', file_path=None, sheet_name='plot_var_bin_summary',
                         need_bin=False, decimal=2, ellipsis=16, **kwargs):
    """
    画变量分箱图
    Args:
        frame (DataFrame):
        cols (str|list): 需要画图的变量
        target (str): 目标变量
        by (str): 分开统计的列名
        file_path (str): 保存路径
        sheet_name (str):保存excel的sheet名称
        need_bin (bool):是否要进行分箱

    Returns:

    """

    if isinstance(cols, str):
        cols = [cols]

    if isinstance(frame, pd.DataFrame):
        summary_df_dict = {}
        if by in frame:
            for name, df in frame.groupby(by):
                summary_df_dict[name] = calc_var_summary(df, include_cols=cols, target=target, need_bin=need_bin,
                                                         **kwargs)
        else:
            summary_df_dict['all'] = calc_var_summary(frame, include_cols=cols, target=target, need_bin=need_bin,
                                                      **kwargs)
    else:
        summary_df_dict = frame
    summary_num = len(summary_df_dict)

    save_jpg_path = None
    if file_path is not None:
        if '.xlsx' in file_path:
            fp_sp = file_path.split('.xlsx')
            if '' == fp_sp[1]:
                save_jpg_path = fp_sp[0] + '_var_jpg'
                os.makedirs(save_jpg_path, exist_ok=True)

            else:
                save_jpg_path = os.path.join(file_path, 'var_jpg')
                os.makedirs(save_jpg_path, exist_ok=True)
                file_path = os.path.join(file_path, 'plot_var_bin_summary.xlsx')
        else:
            save_jpg_path = os.path.join(file_path, 'var_jpg')
            os.makedirs(save_jpg_path, exist_ok=True)
            file_path = os.path.join(file_path, 'plot_var_bin_summary.xlsx')

    for col in cols:
        # 做图
        # plt.figure(figsize=(15, 4))
        # gs = gridspec.GridSpec(1, 2)
        # gs = gridspec.GridSpec(1, 3)

        gs_num = len(summary_df_dict)

        if gs_num == 1:
            plt.figure(figsize=(gs_num * 5, 4), dpi=400)
        else:
            plt.figure(figsize=(gs_num * 5, 4))

        gs = gridspec.GridSpec(1, gs_num)

        for i, k in enumerate(summary_df_dict):

            df = summary_df_dict[k]

            if i == 0:
                ax1 = plt.subplot(gs[0, 0])
            else:
                ax1 = plt.subplot(gs[0, i], sharey=ax1)

            df = df[df['var_name'] == col]
            df['range'] = df['range'].map(lambda x: ellipsis_fun(x, decimal, ellipsis))

            x_point = np.arange(len(df))
            y_point = df['total_pct']

            ax1.bar(x_point, y_point, color='Orange', alpha=0.4, width=0.5, label='PctTotal')
            x_labels = list(df['range'])
            plt.xticks(np.arange(len(df)), x_labels, fontsize=10, rotation=45)

            for x, y in zip(x_point, y_point):
                ax1.text(x + 0.05, y + 0.01, str(round(y * 100, 2)) + '%', ha='center', va='bottom', fontsize=12)
            ax1.set_ylabel('total_pct', fontsize=12)

            ax1.set_ylim([0, max(y_point) + ((max(y_point) - min(y_point)) / len(y_point))])
            bottom, top = ax1.get_ylim()
            ax2 = ax1.twinx()
            ax2.plot(x_point, df['positive_rate'], '-ro', color='red')

            for x, y in zip(x_point, df['positive_rate']):
                ax2.text(x + 0.05, y, str(round(y * 100, 2)) + '%', ha='center', va='bottom', fontsize=12, color='r')
            ax2.set_ylabel('positive_rate', fontsize=12)
            # ax2.set_ylim([0, max(df['positive_rate']) + 0.01])
            ax2.set_ylim(
                [0, max(df['positive_rate']) + ((max(df['positive_rate']) - min(df['positive_rate'])) / len(df))])

            plt.title('{}:{}\nIV: {:.5f}'.format(k, col, df['IV'].iloc[0]), loc='right', fontsize='small')
            # 将数据详情表添加
            tmp_df = df[['range', 'woe', 'iv', 'total']]
            round_cols = ['woe', 'iv']
            tmp_df[round_cols] = tmp_df[round_cols].applymap(lambda v: round(v, 4) if pd.notnull(v) else '')
            mpl_table = plt.table(cellText=tmp_df.values, colLabels=tmp_df.columns,
                                  colWidths=[0.2, 0.1, 0.1, 0.1],
                                  loc='top')  # loc='top'将详情放到顶部

            mpl_table.auto_set_font_size(False)
            mpl_table.set_fontsize(5)

            header_color = 'darkorange'
            row_colors = ['bisque', 'w']
            header_columns = 0

            for k, cell in six.iteritems(mpl_table._cells):
                cell.set_edgecolor('w')
                if k[0] == 0 or k[1] < header_columns:
                    cell.set_text_props(weight='bold', color='w')
                    cell.set_facecolor(header_color)
                else:
                    cell.set_facecolor(row_colors[k[0] % len(row_colors)])

        plt.tight_layout()

        if save_jpg_path is not None:  # 判断是否需要保存
            plt.savefig(os.path.join(save_jpg_path, '{}.png'.format(col)), dpi=300, bbox_inches='tight')  # dpi控制清晰度
        # plt.show()

    # if save_jpg_path is not None:
    #     workbook = xlsxwriter.Workbook(file_path)
    #     worksheet = workbook.add_worksheet(sheet_name)
    #
    #     for i, jpg_name in enumerate(cols):
    #         worksheet.insert_image('A{}'.format(i * 29 + 3), os.path.join(save_jpg_path, '{}.jpg'.format(jpg_name),
    #                                {'x_scale': 1.5, 'y_scale': 1.5})
    #
    #     workbook.close()
    if save_jpg_path is not None:
        try:
            wb = openpyxl.load_workbook(file_path)
        except:
            wb = openpyxl.Workbook()
        sh = wb.create_sheet(sheet_name)
        for i, jpg_name in enumerate(cols):
            img = Image(os.path.join(save_jpg_path, '{}.png'.format(jpg_name)))
            newsize = (summary_num * 700, 600)
            img.width, img.height = newsize  # 设置图片的宽和高
            sh.add_image(img, 'A{}'.format(i * 35 + 3))
        wb.save(file_path)


def plot_describe(df, by_col='p', title='model distribution'):
    try:
        sns.displot(df[by_col], color='#ff8080', bins=20,
                    kde_kws={"lw": 2.5, 'linestyle': '--'})
    except RuntimeError as rte:
        if str(rte).startswith("Selected KDE bandwidth is 0. Cannot estiamte density."):
            sns.displot(df[by_col], color='#ff8080',
                        bins=20, kde_kws={"lw": 2.5, 'linestyle': '--', 'bw': 1})
        else:
            raise rte
    tem = df[by_col].describe().reset_index()
    table_ = plt.table(cellText=[[round(x, 4)] for x in tem[by_col].tolist()],
                       colWidths=[0.1] * 1, rowLabels=tem['index'].tolist(), loc='right')
    table_.set_fontsize(15)
    table_.scale(1.9, 2.265)
    plt.title(f'{title} by {by_col}')


def get_optimal_cutoff(fpr_recall, tpr_precision, threshold, is_f1=False):
    if is_f1:
        youdenJ_f1score = (2 * tpr_precision * fpr_recall) / (tpr_precision + fpr_recall)
    else:
        youdenJ_f1score = tpr_precision - fpr_recall
    point_index = np.argmax(youdenJ_f1score)
    optimal_threshold = threshold[point_index]
    point = [fpr_recall[point_index], tpr_precision[point_index]]
    return optimal_threshold, point


def plot_ks(y_true: pd.Series, y_pred: pd.Series, output_path='', title=''):
    fpr, tpr, thresholds = roc_curve(y_true, y_pred)
    ks = max(tpr - fpr)  ###计算ks的值

    plt.figure(figsize=(6, 6))
    x = np.arange(len(thresholds)) / len(thresholds)
    plt.plot(x, tpr, lw=1)
    plt.plot(x, fpr, lw=1)
    plt.plot(x, tpr - fpr, lw=1, linestyle='--', label='KS curve (KS = %0.4f)' % ks)

    optimal_th, optimal_point = get_optimal_cutoff(fpr_recall=fpr, tpr_precision=tpr, threshold=thresholds)
    optimal_th_index = np.where(thresholds == optimal_th)
    plt.plot(optimal_th_index[0][0] / len(thresholds), ks, marker='o', color='r')
    plt.text(optimal_th_index[0][0] / len(thresholds), ks, (float('%.4f' % optimal_point[0]),
                                                            float('%.4f' % optimal_point[1])),
             ha='right', va='top', fontsize=12)
    plt.text(optimal_th_index[0][0] / len(thresholds), ks, f'threshold:{optimal_th:.4f}', fontsize=12)

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('Thresholds Index')
    plt.ylabel('TPR FPR KS')
    name = '{} KS Curve'.format(title)
    plt.title(name)
    plt.legend(loc="lower right", fontsize=12)
    plt.savefig(output_path + name, bbox_inches='tight')
    plt.show()
    return {'KS': ks, 'KS最大值-threshold': optimal_th}


def plot_roc(y_true: pd.Series, y_pred: pd.Series, output_path='', title=''):
    fpr, tpr, thresholds = roc_curve(y_true, y_pred)
    auc_value = roc_auc_score(y_true, y_pred)  ###计算auc的值

    lw = 2
    plt.figure(figsize=(6, 6))

    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (AUC = %0.4f)' % auc_value)

    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')

    optimal_th, optimal_point = get_optimal_cutoff(fpr_recall=fpr, tpr_precision=tpr, threshold=thresholds)
    plt.plot(optimal_point[0], optimal_point[1], marker='o', color='r')
    plt.text(optimal_point[0], optimal_point[1], (float('%.4f' % optimal_point[0]),
                                                  float('%.4f' % optimal_point[1])),
             ha='right', va='top', fontsize=12)
    plt.text(optimal_point[0], optimal_point[1], f'threshold:{optimal_th:.4f}', fontsize=12)

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('False Positive Rate(FPR)')
    plt.ylabel('True Positive Rate(TPR)')
    name = '{} ROC Curve'.format(title)
    plt.title(name)
    plt.legend(loc="lower right", fontsize=12)
    plt.savefig(output_path + name, bbox_inches='tight')
    plt.show()
    return {'AUC': auc_value}


def plot_pr(y_true: pd.Series, y_pred: pd.Series, output_path='', title=''):
    precision, recall, thresholds = precision_recall_curve(y_true, y_pred)
    f1score = (2 * precision * recall) / (precision + recall)  ###计算F1score
    max_f1score = max(f1score)

    lw = 2
    plt.figure(figsize=(6, 6))

    plt.plot(recall, precision, color='darkorange',
             lw=lw, label='PR curve (F1score = %0.4f)' % max_f1score)

    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')

    optimal_th, optimal_point = get_optimal_cutoff(fpr_recall=recall, tpr_precision=precision, threshold=thresholds,
                                                   is_f1=True)
    plt.plot(optimal_point[0], optimal_point[1], marker='o', color='r')
    plt.text(optimal_point[0], optimal_point[1], (float('%.4f' % optimal_point[0]),
                                                  float('%.4f' % optimal_point[1])),
             ha='right', va='top', fontsize=12)
    plt.text(optimal_point[0], optimal_point[1], f'threshold:{optimal_th:.4f}', fontsize=12)

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    name = '{} PR Curve'.format(title)
    plt.title(name)
    plt.legend(loc="lower right", fontsize=12)
    plt.savefig(output_path + name, bbox_inches='tight')
    plt.show()
    return {'F1_Score最大值': max_f1score, 'F1_Score最大值-threshold': optimal_th, '模型拐点': optimal_th, '阀值': optimal_th,
            'Precision': optimal_point[1], 'Recall': optimal_point[0], 'F1_Score': max_f1score}


def plot_pr_f1(y_true: pd.Series, y_pred: pd.Series, output_path='', title=''):
    precision, recall, thresholds = precision_recall_curve(y_true, y_pred)
    thresholds = np.insert(thresholds, 0, 0, axis=None)
    f1score = (2 * precision * recall) / (precision + recall)  ###计算F1score

    x = np.arange(len(thresholds)) / len(thresholds)

    pr_f1_dict = {'Precision': precision, 'Recall': recall, 'F1_score': f1score}

    for i in pr_f1_dict:
        plt.figure(figsize=(6, 6))

        plt.plot(x, pr_f1_dict[i], lw=1)

        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.0])
        plt.xlabel('Thresholds Index')
        plt.ylabel('{}'.format(i))
        name = '{} {} Curve'.format(title, i)
        plt.title(name)
        plt.savefig(output_path + name, bbox_inches='tight')
        plt.show()

    return {'Thresholds': list(thresholds), '模型召回率': list(recall), '模型精准率': list(precision),
            '模型F1-score': list(f1score)}


def calc_celue_cm(df: pd.DataFrame, target='target', to_bin_col='p'):
    q_cut_list = np.arange(0, 1, 1 / 10) + 0.1
    confusion_matrix_df = pd.DataFrame()
    for i in q_cut_list:
        df['pred_label'] = np.where(df[to_bin_col] >= i, 1, 0)
        tmp_list = []
        tmp_list.append(i)

        tn, fp, fn, tp = confusion_matrix(np.array(df[target]), np.array(df['pred_label'])).ravel()

        tmp_list.extend([tp, fp, tn, fn])

        confusion_matrix_df = confusion_matrix_df.append(pd.DataFrame(tmp_list).T)

    # confusion_matrix_df.columns = ['阈值', 'TP', 'FP', 'TN', 'FN']
    confusion_matrix_df.columns = ['阈值', '实际正样本-预测为正样本', '实际负样本-预测为正样本', '实际负样本-预测为负样本', '实际正样本-预测为负样本']
    confusion_matrix_df.set_index('阈值', inplace=True)
    confusion_matrix_df['sum'] = confusion_matrix_df.apply(lambda x: x.sum(), axis=1)

    # return confusion_matrix_df
    return confusion_matrix_df.to_dict()


def calc_plot_metrics(df: pd.DataFrame, to_bin_col='p', target='target', curve_save_path=''):
    data = {k: v for k, v in df.groupby('type')}
    data.update({'all': df})

    for data_type, type_df in data.items():
        res_save_path = os.path.join(curve_save_path, data_type)
        os.makedirs(res_save_path, exist_ok=True)

        res_dict = {}
        res_dict.update(plot_roc(type_df[target], type_df[to_bin_col], res_save_path, data_type))
        res_dict.update(plot_ks(type_df[target], type_df[to_bin_col], res_save_path, data_type))
        res_dict.update(plot_pr(type_df[target], type_df[to_bin_col], res_save_path, data_type))
        res_dict.update(calc_celue_cm(type_df, target, to_bin_col))
        res_dict.update(plot_pr_f1(type_df[target], type_df[to_bin_col], res_save_path, data_type))

        ###相关指标保存json格式
        save_json(res_dict, os.path.join(res_save_path, '{}_res_json.json'.format(data_type)))


if __name__ == "__main__":
    ######读取数据
    data_path = 'TD47p25combine_td_to_report_data.csv'
    df = pd.read_csv(data_path)
    df = df[df['label'].notnull()]

    print(len(df))

    ######结果保存路径
    curve_save_path = '../examples/curve_result'
    calc_plot_metrics(df=df, to_bin_col='td', target='label', curve_save_path=curve_save_path)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

'''
@File: statistics_mr_auc_ks_test.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2023-12-06
'''

from autobmt.statistics import StatisticsMrAucKs

conf_dict = {
        'cust_id': 'device_id',  # 主键
        'date_na': 'apply_time',  # 时间列名
        'type_na': 'type',  # 数据集列名
        'model_pred_res': 'p',  # 模型预测值
        'target_na': 'target',  # 目标变量列名
        'device_type_na': 'device_type',  # 设备列名
        'year_month_na': 'apply_month',

        ###设备号列名
        'oaid_col_na': 'oaid_md5',
        'imei_col_na': 'imei_md5',
        'idfa_col_na': 'idfa_md5',

        ###设备号的具体取值
        'oaid_value': 'oaid',
        'imei_value': 'imei',
        'idfa_value': 'idfa',
    }

data_path = '/Users/ryanzheng/td/项目/RTA/pddRTA/模型/R8431/TD84p31_oneid_v2_all_rowdrop_to_statistics_mr_auc_ks.csv'
smak = StatisticsMrAucKs(conf_dict, data_path)
smak.statistics_model_mr_auc_ks()




#!/usr/bin/env python
# ! -*- coding: utf-8 -*-


import os
import sys

module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import numpy as np
import pandas as pd

from autobmt.plot import plot_var_bin_summary
from autobmt.transformer import FeatureBin, WoeTransformer, DEFAULT_NAME

np.random.seed(1)

ab = np.array(list('ABCDEFG'))
feature = np.random.randint(10, size=500)
target = np.random.randint(2, size=500)
str_feat = ab[np.random.choice(7, 500)]
uni_feat = np.ones(500)
empty_feat = feature.astype(float)
empty_feat[np.random.choice(500, 50, replace=False)] = np.nan
ty = np.array(['train', 'test', 'oot'])
ty_fea = ty[np.random.choice(3, 500)]

df = pd.DataFrame({
    'A': feature,
    'B': str_feat,
    'C': uni_feat,
    'D': empty_feat,
    'target': target,
    'type': ty_fea,
})


def test_plot_var_bin_summary():
    fb = FeatureBin()
    fb.fit(df.drop(['type'], axis=1), 'target', n_bins=4, is_empty_bin=True, is_need_monotonic=False)
    df_bin = fb.transform(df, labels=True)

    assert df_bin.loc[495, 'B'] == '0.G,A,D'

    file_path = './tests/polt_bin_summary/'
    cols = ['A', 'B', 'C', 'D', 'type']
    plot_var_bin_summary(df_bin, cols, by='type', file_path=file_path, sheet_name='Sheet', need_bin=False)


fb = FeatureBin()
fb.fit(df.drop(['type'], axis=1), 'target', n_bins=4, is_empty_bin=True, is_need_monotonic=False)
df_bin = fb.transform(df, labels=True)

assert df_bin.loc[495, 'B'] == '0.G,A,D'

# file_path = '../tests/polt_bin_summaryttttTruedtatatat/'
# cols = ['A', 'B', 'C', 'D', 'type']
# plot_var_bin_summary(df, cols, by='type', file_path=file_path, sheet_name='Sheet',need_bin=True,is_need_monotonic=True,method='dt')



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: autobmt_treemodel_tutorial_code.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2022-07-11
'''

import pandas as pd

from autobmt.auto_build_tree_model import AutoBuildTreeModel

##**************************************************虚构现实数据例子**************************************************
##**************************************************虚构现实数据例子**************************************************
##**************************************************虚构现实数据例子**************************************************

##**************************************************自动构建lr模型**************************************************

###TODO 注意修改，读取建模数据
data = pd.read_csv('./example_data/tutorial_data.csv')
###TODO 注意修改，读取建模数据

###TODO 注意修改
client_batch = 'TT00p1'
key, target, data_type = 'APP_ID_C', 'target', 'type'  # key是主键字段名，target是目标变量y的字段名，data_type是train、test数据集标识的字段名
ml_res_save_path = './example_model_result/{}'.format(client_batch)  # 模型结果保存的位置
###TODO 注意修改

###TODO 下面代码基本可以不用动
# 初始化
autobtmodel = AutoBuildTreeModel(datasets=data,  # 训练模型的数据集
                                 fea_names=list(data.columns),  # 数据集的字段名
                                 target=target,  # 目标变量y字段名
                                 key=key,  # 主键字段名
                                 data_type=data_type,  # train、test数据集标识的字段名
                                 no_feature_names=[key, target, data_type] + ['apply_time', 'var_d1', 'var_d3',
                                                                              'var_d4', 'var_d5', 'var_d6', 'var_d7',
                                                                              'var_d8', 'var_d11'],
                                 # 数据集中不用于开发模型的特征字段名，即除了x特征的其它字段名
                                 ml_res_save_path=ml_res_save_path,  # 建模相关结果保存路径
                                 AB={'A': 404.65547021957406, 'B': 72.13475204444818},  # 自定义的大A，大B；非必要，有默认的
                                 positive_corr=False,  # 数与模型预测的概率值是否正相关。默认False，负相关，即概率约高，分数越低
                                 )

# 训练模型
model, in_model_fea = autobtmodel.fit(is_feature_select=True,  # 特征筛选
                                      is_auto_tune_params=True,  # 是否自动调参
                                      is_stepwise_del_feature=True,  # 是进行逐步的变量删除
                                      feature_select_method='shap',  # 特征筛选指标
                                      method_threhold=0.001,  # 特征筛选指标阈值
                                      corr_threhold=0.8,  # 相关系数阈值
                                      psi_threhold=0.1,  # PSI阈值
                                      )
###TODO 上面代码基本可以不用动


# ##**************************************************读取未来需要预测的数据，使用训练好的模型进行预测**************************************************
#
# ###TODO 注意修改，读取未来需要预测的数据
# data = pd.read_csv(
#     './example_data/tutorial_data.csv')
# ###TODO 注意修改，读取未来需要预测的数据
#
# # 未来有新数据过来，使用训练好的模型进行预测
# offline_pred_res = AutoBuildTreeModel.predict(to_pred_df=data,  # 未来需要预测的数据，id+特征即可
#                                               model_path='./example_model_result/TT00p1/20231119105732_32',
#                                               # 训练好的模型路径
#                                               )
# offline_pred_res.to_csv(
#     './example_model_result/TT00p1/20231119105732_32/offline_pred_res.csv',
#     index=False)
#
# ##**************************************************虚构现实数据例子**************************************************
# ##**************************************************虚构现实数据例子**************************************************
# ##**************************************************虚构现实数据例子**************************************************



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: auto_build_tree_model.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-11-17
'''

import gc
import json
import os
import time
import warnings

import joblib
import pickle
import numpy as np
import pandas as pd
from xgboost import XGBClassifier
import xgboost as xgb

import autobmt

warnings.filterwarnings('ignore')

from .logger_utils import Logger

log = Logger(level='info', name=__name__).logger


class AutoBuildTreeModel():
    def __init__(self, datasets, fea_names, target, key='key', data_type='type',
                 no_feature_names=['key', 'target', 'apply_time', 'type'], ml_res_save_path='model_result',
                 AB={}, positive_corr=False):

        if data_type not in datasets:
            raise KeyError('train、test数据集标识的字段名不存在！或未进行数据集的划分，请将数据集划分为train、test！！！')

        data_type_ar = np.unique(datasets[data_type])
        if 'train' not in data_type_ar:
            raise KeyError("""没有开发样本，数据集标识字段{}没有`train`该取值！！！""".format(data_type))

        if 'test' not in data_type_ar:
            raise KeyError("""没有验证样本，数据集标识字段{}没有`test`该取值！！！""".format(data_type))

        if target not in datasets:
            raise KeyError('样本中没有目标变量y值！！！')

        # fea_names = [i for i in fea_names if i != key and i != target]
        fea_names = [i for i in fea_names if i not in no_feature_names]
        log.info('数据集变量个数 : {}'.format(len(fea_names)))
        log.info('fea_names is : {}'.format(fea_names))

        self.datasets = datasets
        self.fea_names = fea_names
        self.target = target
        self.key = key
        self.no_feature_names = no_feature_names
        self.ml_res_save_path = os.path.join(ml_res_save_path, time.strftime('%Y%m%d%H%M%S_%S', time.localtime()))
        self.AB = AB
        self.positive_corr = positive_corr  # 分数与模型预测的概率值是否正相关。默认False，负相关，即概率约高，分数越低
        self.min_child_samples = max(round(len(datasets[datasets['type'] == 'train']) * 0.02),
                                     50)  # 一个叶子上数据的最小数量. 可以用来处理过拟合

        os.makedirs(self.ml_res_save_path, exist_ok=True)

    def fit(self, is_feature_select=True, is_auto_tune_params=True, is_stepwise_del_feature=True,
            feature_select_method='shap', method_threhold=0.001,
            corr_threhold=0.8, psi_threhold=0.2):
        '''

        Args:
            is_feature_select:
            is_auto_tune_params:
            feature_select_method:
            method_threhold:
            corr_threhold:
            psi_threhold:

        Returns: xgboost.sklearn.XGBClassifier或lightgbm.sklearn.LGBClassifier；list
            返回最优模型，入模变量list

        '''
        log.info('*' * 30 + '开始自动建模' + '*' * 30)

        log.info('*' * 30 + '获取变量名和数据集' + '*' * 30)
        fea_names = self.fea_names.copy()
        dev_data = self.datasets[self.datasets['type'] == 'train']
        nodev_data = self.datasets[self.datasets['type'] == 'test']

        del self.datasets;
        gc.collect()

        # dev_data = self.datasets['dev']
        # nodev_data = self.datasets['nodev']

        params = {
            'learning_rate': 0.05,
            'n_estimators': 200,
            'max_depth': 3,
            #'min_child_weight': max(round(len(dev_data) * 0.01), 50),
            'min_child_weight': 5,
            'gamma': 7,
            'subsample': 0.7,
            'colsample_bytree': 0.9,
            'colsample_bylevel': 0.7,
            'reg_alpha': 10,
            'reg_lambda': 10,
            'scale_pos_weight': 1
        }
        log.info('默认参数 {}'.format(params))

        best_model = XGBClassifier(**params)
        # best_model.fit(dev_data[fea_names], dev_data[self.target])
        log.info('构建基础模型')

        if is_feature_select:
            log.info('需要进行变量筛选')
            fea_names = autobmt.feature_select({'dev': dev_data, 'nodev': nodev_data}, fea_names, self.target,
                                               feature_select_method, method_threhold,
                                               corr_threhold,
                                               psi_threhold)

        if is_auto_tune_params:
            log.info('需要进行自动调参')
            best_model = autobmt.classifiers_model_auto_tune_params(
                train_data=(dev_data[fea_names], dev_data[self.target]),
                test_data=(nodev_data[fea_names], nodev_data[self.target]))
            params = best_model.get_params()

        if is_stepwise_del_feature:
            log.info('需要逐步的删除变量')
            _, fea_names = autobmt.stepwise_del_feature(best_model, {'dev': dev_data, 'nodev': nodev_data}, fea_names,
                                                        self.target,
                                                        params)

        # 最终模型
        log.info('使用自动调参选出来的最优参数+筛选出来的变量，构建最终模型')
        log.info('最终变量的个数{}, 最终变量{}'.format(len(fea_names), fea_names))
        log.info('自动调参选出来的最优参数{}'.format(params))
        # xgb_clf = XGBClassifier(**params)
        # xgb_clf.fit(dev_data[fea_names], dev_data[self.target])
        best_model.fit(dev_data[fea_names], dev_data[self.target])

        # ###
        # pred_nodev = xgb_clf.predict_proba(nodev_data[fea_names])[:, 1]
        # pred_dev = xgb_clf.predict_proba(dev_data[fea_names])[:, 1]
        # df_pred_nodev = pd.DataFrame({'target': nodev_data[self.target], 'p': pred_nodev}, index=nodev_data.index)
        # df_pred_dev = pd.DataFrame({'target': dev_data[self.target], 'p': pred_dev}, index=dev_data.index)
        # ###

        # ###
        # df_pred_nodev = nodev_data[self.no_feature_names + fea_names]
        # df_pred_dev = dev_data[self.no_feature_names + fea_names]
        # df_pred_nodev['p'] = xgb_clf.predict_proba(df_pred_nodev[fea_names])[:, 1]
        # df_pred_dev['p'] = xgb_clf.predict_proba(df_pred_dev[fea_names])[:, 1]
        # ###

        ###
        df_pred_nodev = nodev_data[self.no_feature_names]
        df_pred_dev = dev_data[self.no_feature_names]
        df_pred_nodev['p'] = best_model.predict_proba(nodev_data[fea_names])[:, 1]
        df_pred_dev['p'] = best_model.predict_proba(dev_data[fea_names])[:, 1]
        ###

        # 计算auc、ks、psi
        test_ks = autobmt.get_ks(df_pred_nodev[self.target], df_pred_nodev['p'])
        train_ks = autobmt.get_ks(df_pred_dev[self.target], df_pred_dev['p'])
        test_auc = autobmt.get_auc(df_pred_nodev[self.target], df_pred_nodev['p'])
        train_auc = autobmt.get_auc(df_pred_dev[self.target], df_pred_dev['p'])

        q_cut_list = np.arange(0, 1, 1 / 20)
        bins = np.append(np.unique(np.quantile(df_pred_nodev['p'], q_cut_list)), df_pred_nodev['p'].max() + 0.1)
        df_pred_nodev['range'] = pd.cut(df_pred_nodev['p'], bins=bins, precision=0, right=False).astype(str)
        df_pred_dev['range'] = pd.cut(df_pred_dev['p'], bins=bins, precision=0, right=False).astype(str)
        nodev_psi = autobmt.psi(df_pred_nodev['range'], df_pred_dev['range'])
        res_dict = {'dev_auc': train_auc, 'nodev_auc': test_auc, 'dev_ks': train_ks, 'nodev_ks': test_ks,
                    'nodev_dev_psi': nodev_psi}
        log.info('auc & ks & psi: {}'.format(res_dict))
        log.info('*' * 30 + '自动构建模型完成！！！' + '*' * 30)

        ##############
        log.info('*' * 30 + '建模相关结果开始保存！！！' + '*' * 30)
        joblib.dump(best_model._Booster, os.path.join(self.ml_res_save_path, 'xgb.ml'))
        joblib.dump(best_model, os.path.join(self.ml_res_save_path, 'xgb_sk.ml'))
        autobmt.dump_to_pkl(best_model._Booster, os.path.join(self.ml_res_save_path, 'xgb.pkl'))
        autobmt.dump_to_pkl(best_model, os.path.join(self.ml_res_save_path, 'xgb_sk.pkl'))
        json.dump(best_model.get_params(), open(os.path.join(self.ml_res_save_path, 'xgb.params'), 'w'))
        best_model._Booster.dump_model(os.path.join(self.ml_res_save_path, 'xgb.txt'))
        pd.DataFrame([res_dict]).to_csv(os.path.join(self.ml_res_save_path, 'xgb_auc_ks_psi.csv'), index=False)

        pd.DataFrame(list(best_model._Booster.get_fscore().items()),
                     columns=['fea_names', 'weight']
                     ).sort_values('weight', ascending=False).set_index('fea_names').to_csv(
            os.path.join(self.ml_res_save_path, 'xgb_featureimportance.csv'))

        nodev_data[self.no_feature_names + fea_names].head(500).to_csv(
            os.path.join(self.ml_res_save_path, 'xgb_test_input.csv'),
            index=False)

        ##############pred to score
        df_pred_nodev['score'] = df_pred_nodev['p'].map(
            lambda x: autobmt.to_score(x, self.AB['A'], self.AB['B'], self.positive_corr))
        df_pred_dev['score'] = df_pred_dev['p'].map(
            lambda x: autobmt.to_score(x, self.AB['A'], self.AB['B'], self.positive_corr))
        ##############pred to score

        df_pred_nodev.append(df_pred_dev).to_csv(os.path.join(self.ml_res_save_path, 'xgb_pred_to_report_data.csv'),
                                                 index=False)

        log.info('*' * 30 + '建模相关结果保存完成！！！保存路径为：{}'.format(self.ml_res_save_path) + '*' * 30)

        return best_model, fea_names

    @classmethod
    def predict(cls, to_pred_df=None, model_path=None):
        if to_pred_df is None:
            raise ValueError('需要进行预测的数据集不能为None，请指定数据集！！！')
        if model_path is None:
            raise ValueError('模型路径不能为None，请指定模型文件路径！！！')

        try:
            model = joblib.load(os.path.join(model_path, 'xgb.ml'))
        except:
            model = pickle.load(open(os.path.join(model_path, 'xgb.pkl'), 'rb'))

        try:
            model_feature_names = model.feature_names
        except:
            model_feature_names = model.get_booster().feature_names
            model = model.get_booster()

        to_pred_df['p'] = model.predict(xgb.DMatrix(to_pred_df[model_feature_names]))

        return to_pred_df



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: bayes_opt_tuner.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-04-25
'''

import time
import warnings

import numpy as np
import pandas as pd
from bayes_opt import BayesianOptimization
from lightgbm import LGBMClassifier
from xgboost import XGBClassifier

from .metrics import get_ks, get_auc
from .utils import get_accuracy, get_recall, get_precision, get_f1, r2

warnings.filterwarnings('ignore')

from .logger_utils import Logger

log = Logger(level='info', name=__name__).logger


class ModelTune():
    def __init__(self):
        self.base_model = None
        self.best_model = None
        self.tune_params = None
        self.loss = np.inf
        self.default_params = None
        self.int_params = None
        self.init_params = None
        self.metrics = None
        self._metrics_score = []
        self.scores = []

    def get_model(self):
        return self.best_model

    def _map_metrics(self):
        mapper = {
            'accuracy': get_accuracy,
            'f1': get_f1,
            'precision': get_precision,
            'recall': get_recall,
            'r2': r2,
            'auc': get_auc,
            'ks': get_ks
        }

        for metric in self.metrics:
            if metric not in mapper:
                raise ValueError('指定的指标 ''`{}` 不支持'.format(metric))
            self._metrics_score.append(mapper[metric])
            self.scores.append(0.)

    def fit(self, train_data=(), test_data=()
            , init_points=30, iterations=120, metrics=[]):
        '''

        Args:
            train_data:
            test_data:
            init_points:
            iterations:
            metrics:

        Returns:

        '''

        if len(metrics) > 0:
            self.metrics = metrics
        self._map_metrics()

        X_train, y_train = train_data
        X_test, y_test = test_data

        def loss_fun(train_result, test_result, weight=0.3):

            # return test_result - 2 ** abs(test_result - train_result)
            return test_result - 2 ** abs(test_result - train_result) * weight

        # def loss_fun(train_result, test_result):
        #     train_result = train_result * 100
        #     test_result = test_result * 100
        #
        #     return train_result - 2 ** abs(train_result - test_result)

        def obj_fun(**params):
            for param in self.int_params:
                params[param] = int(round(params[param]))

            model = self.base_model(**params, **self.default_params)
            model.fit(X_train, y_train)

            pred_test = model.predict_proba(X_test)[:, 1]
            pred_train = model.predict_proba(X_train)[:, 1]

            # test_auc = get_auc(y_test, pred_test)
            # train_auc = get_auc(y_train, pred_train)
            # print('test_auc is : ', test_auc)
            # print('train_auc is : ', train_auc)

            test_ks = get_ks(y_test, pred_test)
            train_ks = get_ks(y_train, pred_train)
            # print('test_ks is : ', test_ks)
            # print('train_ks is : ', train_ks)

            # maximize = loss_fun(train_auc, test_auc)
            maximize = loss_fun(train_ks, test_ks)
            # print('max_result is : ', maximize)
            # max_result = loss_fun(train_ks, test_ks) * 2 + loss_fun(train_auc, test_auc)

            loss = -maximize
            if loss < self.loss:
                self.loss = loss
                self.best_model = model
                # print('best model result is {}'.format(loss))
                # print('best model params is : ')
                # print(self.best_model.get_params())
                for i, _metric in enumerate(self._metrics_score):
                    self.scores[i] = _metric(y_test, pred_test)
            # print('current obj_fun result is : ', maximize)

            return maximize

        params_optimizer = BayesianOptimization(obj_fun, self.tune_params, random_state=1)
        log.info('需要优化的超参数是 : {}'.format(params_optimizer.space.keys))

        log.info('开始优化超参数!!!')
        start = time.time()

        params_optimizer.maximize(init_points=1, n_iter=0, acq='ei',
                                  xi=0.0)
        params_optimizer.probe(self.init_params, lazy=True)

        # params_optimizer.probe(self.init_params, lazy=True)

        params_optimizer.maximize(init_points=0, n_iter=0)

        params_optimizer.maximize(init_points=init_points, n_iter=iterations, acq='ei',
                                  xi=0.0)  # init_points：探索开始探索之前的迭代次数；iterations：方法试图找到最大值的迭代次数
        # params_optimizer.maximize(init_points=init_points, n_iter=iterations, acq='ucb', xi=0.0, alpha=1e-6)
        end = time.time()
        log.info('优化参数结束!!! 共耗时{} 分钟'.format((end - start) / 60))
        log.info('最优参数是 : {}'.format(params_optimizer.max['params']))
        log.info('{} model 最大化的结果 : {}'.format(type(self.best_model), params_optimizer.max['target']))


class ClassifierModel(ModelTune):
    def __init__(self):
        super().__init__()
        self.metrics = ['auc', 'ks']


class RegressorModel(ModelTune):
    def __init__(self):
        super().__init__()
        self.metrics = ['r2', 'rmse']


class XGBClassifierTuner(ClassifierModel):
    def __init__(self):
        super().__init__()

        self.base_model = XGBClassifier
        self.tune_params = {
            'learning_rate': (0.01, 0.15),
            'n_estimators': (90, 300),
            'max_depth': (2, 7),
            'min_child_weight': (1, 300),
            'subsample': (0.4, 1.0),
            'colsample_bytree': (0.3, 1.0),
            'colsample_bylevel': (0.5, 1.0),
            'gamma': (0, 20.0),
            'reg_alpha': (0, 20.0),
            'reg_lambda': (0, 20.0),
            # 'scale_pos_weight': (1, 5),
            # 'max_delta_step': (0, 10)
        }

        self.default_params = {
            'objective': 'binary:logistic',
            'n_jobs': -1,
            'nthread': -1
        }

        self.init_params = {
            'learning_rate': 0.05,
            'n_estimators': 200,
            'max_depth': 3,
            'min_child_weight': 5,
            'subsample': 0.7,
            'colsample_bytree': 0.9,
            'colsample_bylevel': 0.7,
            'gamma': 7,
            'reg_alpha': 10,
            'reg_lambda': 10,
            # 'scale_pos_weight': 1
        }

        self.int_params = ['max_depth', 'n_estimators']


class LGBClassifierTuner(ClassifierModel):
    '''
    英文版：https://lightgbm.readthedocs.io/en/latest/Parameters.html

    中文版：https://lightgbm.apachecn.org/#/docs/6

    其他注解：https://medium.com/@gabrieltsen
    '''

    def __init__(self):
        super().__init__()

        self.base_model = LGBMClassifier
        self.tune_params = {
            'max_depth': (3, 15),
            'num_leaves': (16, 128),
            'learning_rate': (0.01, 0.2),
            'reg_alpha': (0, 100),
            'reg_lambda': (0, 100),
            'min_child_samples': (1, 100),
            'min_child_weight': (0.01, 100),
            'colsample_bytree': (0.5, 1),
            'subsample': (0.5, 1),
            'subsample_freq': (2, 10),
            'n_estimators': (90, 500),

        }

        self.default_params = {
            'boosting_type': 'gbdt',
            'objective': 'binary',
            'metric': 'auc',
            'random_state': 1024,
            'n_jobs': -1,
            'num_threads': -1,
            'verbose': -1,

        }

        self.init_params = {
            'max_depth': -1,
            "num_leaves": 31,
            "learning_rate": 0.02,
            "reg_alpha": 0.85,
            "reg_lambda": 3,
            "min_child_samples": 20,  # TODO注意修改 .......sklearn:min_child_samples    原生:min_data、min_data_in_leaf
            "min_child_weight": 0.05,  # sklearn:min_child_weight    原生:min_hessian、min_sum_hessian_in_leaf
            "colsample_bytree": 0.9,  # sklearn:colsample_bytree   原生:feature_fraction
            "subsample": 0.8,  # sklearn:subsample  原生:bagging_fraction
            "subsample_freq": 2,  # sklearn:subsample_freq  原生:bagging_freq
            "n_estimators": 100  # sklearn:n_estimators  原生:num_boost_round、num_iterations
        }

        self.int_params = ['max_depth', 'num_leaves', 'min_child_samples', 'n_estimators', 'subsample_freq']


classifiers_dic = {
    # 'lr': LogisticRegressionTuner,
    # 'rf': RandomForestClassifierTuner,
    'xgb': XGBClassifierTuner,
    'lgb': LGBClassifierTuner
}


def classifiers_model_auto_tune_params(models=['xgb'], metrics=[], train_data=(), test_data=()
                                       , init_points=30, iterations=120, verbose=1):
    '''

    Args:
        models:
        metrics:
        train_data:
        test_data:
        init_points:
        iterations:
        verbose:

    Returns:

    '''
    best_model = None
    if not isinstance(models, list):
        raise AttributeError('models参数必须是一个列表, ', '但实际是 {}'.format(type(models)))
    if len(models) == 0:
        models = list(classifiers_dic.keys())
    classifiers = []
    for model in models:
        if model in classifiers_dic:
            classifiers.append(classifiers_dic[model])
    loss = np.inf
    _model = None
    for classifier in classifiers:
        if verbose:
            log.info("优化 {}...".format(classifier()))
        _model = classifier()
        _model.fit(train_data=train_data,
                   test_data=test_data
                   , init_points=init_points, iterations=iterations, metrics=metrics)

        _loss = _model.loss
        if verbose:
            _show_fit_log(_model)
        if _loss < loss:
            loss = _loss
            best_model = _model

    return best_model.get_model()


def _show_fit_log(model):
    _out = '  最优结果: '
    _out += ' loss: {:.3}'.format(model.loss)
    _out += ' 测试集 '
    for i, _metric in enumerate(model.metrics):
        _out += ' {}: {:.3}'.format(_metric[:3],
                                    model.scores[i])
    log.info(_out)


if __name__ == '__main__':
    X = pd.read_pickle('X_train.pkl')
    X = pd.DataFrame(X)
    y = pd.read_pickle('y_train.pkl')
    y = pd.Series(y)
    X_test = pd.read_pickle('X_test.pkl')
    X_test = pd.DataFrame(X_test)
    y_test = pd.read_pickle('y_test.pkl')
    y_test = pd.Series(y_test)

    ####build model
    best_model = classifiers_model_auto_tune_params(train_data=(X, y), test_data=(X_test, y_test), verbose=1,
                                                    init_points=1,
                                                    iterations=2)
    # best_model = classifiers_model_auto_tune_params(train_data=(X, y), test_data=(X_test, y_test), verbose=1)
    print('classifiers_model run over!!!')
    print(type(best_model))
    print(best_model.get_params())
    train_pred_y = best_model.predict_proba(X)[:, 1]
    test_pred_y = best_model.predict_proba(X_test)[:, 1]
    ####build model

    #####build model
    # best_model = LGBMClassifier()
    # best_model.fit(X,y)
    # print(best_model.get_params())
    # train_pred_y = best_model.predict_proba(X)[:, 1]
    # test_pred_y = best_model.predict_proba(X_test)[:, 1]
    #####build model

    # #####build model
    # import lightgbm as lgb
    #
    # init_params = {
    #     "boosting_type": "gbdt",
    #     "objective": "binary",
    #     "metric": "auc",
    # }
    # best_model = lgb.train(params=init_params, train_set=lgb.Dataset(X, y), valid_sets=lgb.Dataset(X_test, y_test))
    # best_model.save_model('lgb.txt')
    # json_model = best_model.dump_model()
    # import json
    #
    # with open('lgb.json', 'w') as f:
    #     json.dump(json_model, f)
    # train_pred_y = best_model.predict(X)
    # test_pred_y = best_model.predict(X_test)
    # #####build model

    train_auc = get_auc(y, train_pred_y)
    test_auc = get_auc(y_test, test_pred_y)
    train_ks = get_ks(y, train_pred_y)
    test_ks = get_ks(y_test, test_pred_y)
    print('train_auc is : ', train_auc, 'test_auc is : ', test_auc)
    print('train_ks is : ', train_ks, 'test_ks is : ', test_ks)

    # # #####build model
    # params = {
    #     'learning_rate': 0.05,
    #     'n_estimators': 200,
    #     'max_depth': 3,
    #     'min_child_weight': 5,
    #     'gamma': 7,
    #     'subsample': 0.7,
    #     'colsample_bytree': 0.9,
    #     'colsample_bylevel': 0.7,
    #     'reg_alpha': 10,
    #     'reg_lambda': 10,
    #     'scale_pos_weight': 1
    # }
    #
    # clf = XGBClassifier(**params)
    # clf.fit(X, y)
    # estimator = clf.get_booster()
    # temp = estimator.save_raw()[4:]
    # # #####build model

    ####构建数据
    # from sklearn.datasets import make_classification
    # from sklearn.model_selection import train_test_split
    # import pickle
    # X, y = make_classification(n_samples=10000, random_state=1024)
    # X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)
    #
    # with open('X_train.pkl', 'wb') as f:
    #     f.write(pickle.dumps(X_train))
    # with open('y_train.pkl', 'wb') as f:
    #     f.write(pickle.dumps(y_train))
    # with open('X_test.pkl', 'wb') as f:
    #     f.write(pickle.dumps(X_test))
    # with open('y_test.pkl', 'wb') as f:
    #     f.write(pickle.dumps(y_test))



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

import os
import sys
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)

import pytest
import numpy as np
import pandas as pd

from autobmt.transformer import FeatureBin, WoeTransformer, DEFAULT_NAME

np.random.seed(1)

ab = np.array(list('ABCDEFG'))
feature = np.random.randint(10, size=500)
target = np.random.randint(2, size=500)
str_feat = ab[np.random.choice(7, 500)]
uni_feat = np.ones(500)
empty_feat = feature.astype(float)
empty_feat[np.random.choice(500, 50, replace=False)] = np.nan

df = pd.DataFrame({
    'A': feature,
    'B': str_feat,
    'C': uni_feat,
    'D': empty_feat,
    'target': target,
})


def test_duplicated_keys():
    dup_df = df.rename(columns={"C": "A"})
    with pytest.raises(Exception, match=r"X has duplicate keys `.*`"):
        WoeTransformer().fit_transform(dup_df, target)


def test_woe_transformer():
    f = WoeTransformer().fit_transform(feature, target)
    assert f[451] == pytest.approx(-0.17061154127869285)


def test_woe_transformer_with_str():
    f = WoeTransformer().fit_transform(str_feat, target)
    assert f[451] == pytest.approx(-0.2198594761130199)


def test_woe_transformer_with_unknown_group():
    transer = WoeTransformer().fit(str_feat, target)
    res = transer.transform(['Z'], default='min')
    assert res[0] == pytest.approx(-0.2198594761130199)


def test_woe_transformer_frame():
    res = WoeTransformer().fit_transform(df, target)
    assert res.iloc[451, 1] == pytest.approx(-0.2198594761130199)


def test_woe_transformer_dict():
    transer = WoeTransformer().fit(df, 'target')
    res = transer.transform({
        "A": 6,
        "B": "C",
        "C": 1,
        "D": 2,
    })
    assert res['B'].item() == pytest.approx(-0.09149433112609942)


def test_woe_transformer_select_dtypes():
    res = WoeTransformer().fit_transform(df, target, select_dtypes='object')
    assert res.loc[451, 'A'] == 3


def test_woe_transformer_exclude():
    res = WoeTransformer().fit_transform(df, target, exclude='A')
    assert res.loc[451, 'A'] == 3


def test_woe_transformer_export_single():
    transer = WoeTransformer().fit(feature, target)
    t = transer.export()
    assert t[DEFAULT_NAME][5] == pytest.approx(0.3938235330926786)


def test_woe_transformer_export():
    transer = WoeTransformer().fit(df, target)
    t = transer.export()
    assert t['C'][1] == 0


def test_woe_transformer_load():
    rules = {
        'A': {
            1: 0.1,
            2: 0.2,
            3: 0.3,
        }
    }

    transer = WoeTransformer().load(rules)
    assert transer.fea_woe_dict['A'][2] == 0.2


def test_featurebin():
    f = FeatureBin().fit_transform(feature, target, method='chi', is_need_monotonic=False)
    assert f[451] == 3


def test_featurebin_with_str():
    f = FeatureBin().fit_transform(str_feat, target, method='chi', is_need_monotonic=False)
    assert f[451] == 0


def test_featurebin_unique_feature():
    f = FeatureBin().fit_transform(uni_feat, target, method='chi', is_need_monotonic=False)
    assert f[451] == 0


def test_featurebin_frame():
    res = FeatureBin().fit_transform(df, target)
    assert res.iloc[404, 1] == 2


def test_featurebin_select_dtypes():
    res = FeatureBin().fit_transform(df, target, select_dtypes='number')
    assert res.loc[451, 'B'] == 'G'


def test_featurebin_exclude():
    res = FeatureBin().fit_transform(df, target, exclude='B')
    assert res.loc[451, 'B'] == 'G'


def test_featurebin_labels():
    fb = FeatureBin().fit(df, target, is_need_monotonic=False)
    res = fb.transform(df, labels=True)
    assert res.loc[451, 'A'] == '3.[3 ~ 4)'


# def test_featurebin_single_feature():
#     fb = FeatureBin().fit(df['A'], method = 'step', n_bins = 5)
#     res = fb.transform(df['A'])
#     assert res[451] == 1

def test_featurebin_export():
    fb = FeatureBin().fit(df, target, method='chi', n_bins=4, is_need_monotonic=False)
    bins = fb.export(bin_format=False)
    assert isinstance(bins['B'][0], list)


def test_featurebin_update():
    fb = FeatureBin().fit(df, target, method='chi', n_bins=4, is_need_monotonic=False)
    fb.manual_bin({'A': [1, 2, 3, 4, 5, 6]})
    bins = fb.export(bin_format=False)
    assert len(bins['A']) == 6


# def test_featurebin_step():
#     fb = FeatureBin().fit(df['A'], method = 'step', n_bins = 4)
#     bins = fb.export()
#     assert bins['A'][1] == 4.5

def test_featurebin_target_in_frame():
    fb = FeatureBin().fit(df, 'target', n_bins=4, is_need_monotonic=False)
    bins = fb.export(bin_format=False)
    assert bins['A'][1] == 6


def test_featurebin_target_in_frame_kwargs():
    fb = FeatureBin().fit(df, y='target', n_bins=4, is_need_monotonic=False)
    bins = fb.export(bin_format=False)
    assert bins['A'][1] == 6


def test_featurebin_empty_separate():
    fb = FeatureBin()
    bins = fb.fit_transform(df, 'target', n_bins=4, is_empty_bin=True, is_need_monotonic=False)
    mask = pd.isna(df['D'])
    assert (bins['D'][~mask] != 4).all()


def test_featurebin_labels_with_empty():
    fb = FeatureBin().fit(df, 'target', n_bins=4, is_empty_bin=True, is_need_monotonic=False)
    res = fb.transform(df, labels=True)
    assert res.loc[2, 'D'] == '4.nan'



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: feature_binning.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2020-05-20
'''

import warnings

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.tree import DecisionTreeClassifier, _tree

from .statistics import calc_iv_and_inflection_point, bin_badrate
from .utils import fillna, split_empty, split_points_to_bin, t_sum_np, t_min_np, \
    t_cols_sum_axis_0_np, t_cols_sum_axis_1_np, support_dataframe

warnings.filterwarnings('ignore')

from .logger_utils import Logger

log = Logger(level="info", name=__name__).logger

DEFAULT_BINS = 10
IS_EMPTY_BIN_DEFAULT_BINS = 9


def equal_freq_bin(feature, target=None, min_sample_rate=0.05,
                   n_bins=None,
                   q_cut_list=None, is_need_monotonic=True, is_empty_bin=True):
    """
    等频分箱

    Args:
        feature (array) : 某个x特征
        target (array) : 目标值y变量
        min_sample_rate (number) : 每个箱子的最小占比
        n_bins (int) : 需要分成几箱
        q_cut_list (array、list) : 百分比分割点列表
        is_need_monotonic (bool) : 是否强制单调
        is_empty_bin (bool) : 是否将空箱单独归为一个箱子

    Returns:
        array (array): 分割点
    """

    empty_mask = np.array([])
    if is_empty_bin:
        if target is None:
            feature, empty_mask = split_empty(feature, target)
        else:
            feature, target, empty_mask = split_empty(feature, target)
    else:
        feature = fillna(feature)

    if n_bins is None and q_cut_list is None and is_empty_bin and empty_mask.any():
        n_bins = IS_EMPTY_BIN_DEFAULT_BINS
    elif n_bins is None and q_cut_list is None:
        n_bins = DEFAULT_BINS

    if q_cut_list is None:
        q_cut_list = np.arange(0, 1, 1 / n_bins)

    is_monotonic = False
    while not is_monotonic:
        splits_tmp = np.quantile(feature, q_cut_list)
        splits = np.unique(splits_tmp)[1:]

        x_bins = split_points_to_bin(feature, splits)  # 返回0、1、2、3这样的数值分箱
        if target is None:
            _ = bin_badrate(x_bins)
            is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足
        else:
            bin_badrate_li, _ = bin_badrate(x_bins, target)
            is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足
            if is_need_monotonic:
                if is_monotonic:  # 满足最小分箱占比后，判断是否满足单调性
                    is_monotonic = pd.Series(bin_badrate_li).is_monotonic_decreasing or pd.Series(
                        bin_badrate_li).is_monotonic_increasing

        if n_bins <= 2:
            break

        n_bins = q_cut_list.size - 1  # 这种是从初始的n_bins(10)，一点一点的减
        q_cut_list = np.arange(0, 1, 1 / n_bins)

    if is_empty_bin and empty_mask.any():
        splits = np.append(splits, np.nan)

    return splits


def dt_bin(feature, target, min_sample_rate=0.05, n_bins=None,
           is_need_monotonic=True, is_empty_bin=True, **kwargs):
    """
    决策树分箱

    Args:
        feature (array) : 某个x特征
        target (array) : 目标值y变量
        min_sample_rate (number) : 每个箱子的最小占比
        n_bins (int) : 需要分成几箱
        is_need_monotonic (bool) : 是否强制单调
        is_empty_bin (bool) : 是否将空箱单独归为一个箱子
        **kwargs : 决策树的其它参数

    Returns:
        array (array): 分割点
    """

    empty_mask = np.array([])
    if is_empty_bin:
        feature, target, empty_mask = split_empty(feature, target)
    else:
        feature = fillna(feature)

    if n_bins is None and is_empty_bin and empty_mask.any():
        n_bins = IS_EMPTY_BIN_DEFAULT_BINS
    elif n_bins is None:
        n_bins = DEFAULT_BINS

    is_monotonic = False
    while not is_monotonic:

        # 决策树分箱逻辑
        '''
        1、初始n_b：10（point），splits：9（point）
        2、9+3=12（point） 》11（bin）
        3、11（bin）- 1 ---》10（bin）
        4、老版本有，新版本跳过
        5、10 - 1 = 9（point）
        '''
        tree = DecisionTreeClassifier(
            min_samples_leaf=min_sample_rate,
            max_leaf_nodes=n_bins,
            # 优先满足min_samples_leaf参数。在满足min_samples_leaf参数参数后，再考虑max_leaf_nodes。
            # 比如情况1：min_samples_leaf设置成0.05，max_leaf_nodes设置成20。满足0.05后，最大max_leaf_nodes只有10，那也就这样了
            # 比如情况2：min_samples_leaf设置成0.05，max_leaf_nodes设置成6。满足0.05后，最大max_leaf_nodes有10，那再考虑max_leaf_nodes，继续分到满足max_leaf_nodes=6停止
            # ps:min_samples_leaf=1表示没有限制
            **kwargs,
        )
        tree.fit(feature.reshape((-1, 1)), target)
        thresholds = tree.tree_.threshold
        thresholds = thresholds[thresholds != _tree.TREE_UNDEFINED]
        splits = np.sort(thresholds)
        # 决策树分箱逻辑

        is_monotonic = True
        if is_need_monotonic:
            x_bins = split_points_to_bin(feature, splits)  # 返回0、1、2、3这样的数值分箱

            bin_badrate_li, _ = bin_badrate(x_bins, target)

            # 不需要判断，tree里面已经判断了
            # is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足

            is_monotonic = pd.Series(bin_badrate_li).is_monotonic_decreasing or pd.Series(
                bin_badrate_li).is_monotonic_increasing

        n_bins = len(splits) + 1  # 初始n_bins为10，对应的splits为9。需要+1，在后面的n_bins -= 1后，n_bins才会是9

        if n_bins <= 2:
            break
        n_bins -= 1

    if is_empty_bin and empty_mask.any():
        splits = np.append(splits, np.nan)
    return splits


def kmeans_bin(feature, target=None, min_sample_rate=0.05,
               n_bins=None,
               random_state=1, is_need_monotonic=True, is_empty_bin=True):
    """
    kmeans聚类分箱

    Args:
        feature (array) : 某个x特征
        target (array) :  目标值y变量
        min_sample_rate (number) : 每个箱子的最小占比
        n_bins (int): 需要分成几箱
        random_state (int): kmeans模型中的随机种子
        is_need_monotonic (bool) : 是否强制单调
        is_empty_bin (bool) : 是否将空箱单独归为一个箱子

    Returns:
        array (array): 分割点
    """

    empty_mask = np.array([])
    if is_empty_bin:
        if target is None:
            feature, empty_mask = split_empty(feature, target)
        else:
            feature, target, empty_mask = split_empty(feature, target)
    else:
        feature = fillna(feature)

    if n_bins is None and is_empty_bin and empty_mask.any():
        n_bins = IS_EMPTY_BIN_DEFAULT_BINS
    elif n_bins is None:
        n_bins = DEFAULT_BINS

    is_monotonic = False
    while not is_monotonic:

        # kmeans 逻辑
        kmeans = KMeans(
            n_clusters=n_bins,
            random_state=random_state
        )
        kmeans.fit(feature.reshape((-1, 1)), target)

        centers = np.sort(kmeans.cluster_centers_.reshape(-1))

        l = len(centers) - 1
        splits = np.zeros(l)
        for i in range(l):
            splits[i] = (centers[i] + centers[i + 1]) / 2
        # kmeans 逻辑

        x_bins = split_points_to_bin(feature, splits)  # 返回0、1、2、3这样的数值分箱
        if target is None:
            _ = bin_badrate(x_bins)
            is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足
        else:
            bin_badrate_li, _ = bin_badrate(x_bins, target)
            is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足
            if is_need_monotonic:
                if is_monotonic:  # 满足最小分箱占比后，判断是否满足单调性
                    is_monotonic = pd.Series(bin_badrate_li).is_monotonic_decreasing or pd.Series(
                        bin_badrate_li).is_monotonic_increasing

        n_bins = len(splits) + 1  # 初始n_bins为10，对应的splits为9。需要+1，在后面的n_bins -= 1后，n_bins才会是9

        if n_bins <= 2:
            break
        n_bins -= 1

    if is_empty_bin and empty_mask.any():
        splits = np.append(splits, np.nan)

    return splits


def chi_bin(feature, target, balance=True, min_sample_rate=0.05, n_bins=None,
            is_need_monotonic=True, is_empty_bin=True, min_threshold=None):
    """
    卡方分箱

    Args:
        feature (array) : 某个x特征
        target (array) :  目标值y变量
        min_sample_rate (number) : 每个箱子的最小占比
        n_bins (int): 需要分成几箱
        is_need_monotonic (bool) : 是否强制单调
        is_empty_bin (bool) : 是否将空箱单独归为一个箱子
        min_threshold (number): 最小的卡方阀值

    Returns:
        array (array): 分割点
    """

    empty_mask = np.array([])
    if is_empty_bin:
        feature, target, empty_mask = split_empty(feature, target)
    else:
        feature = fillna(feature)

    if n_bins is None and min_threshold is None and is_empty_bin and empty_mask.any():
        n_bins = IS_EMPTY_BIN_DEFAULT_BINS
    elif n_bins is None and min_threshold is None:
        n_bins = DEFAULT_BINS

    if min_sample_rate and min_sample_rate < 1:
        min_sample_rate = len(feature) * min_sample_rate

    ###
    target_unique = np.unique(target)
    feature_unique = np.unique(feature)
    len_f = len(feature_unique)
    len_t = len(target_unique)

    grouped = np.zeros((len_f, len_t), dtype=float)

    for r in range(len_f):
        tmp = target[feature == feature_unique[r]]
        for c in range(len_t):
            grouped[r, c] = (tmp == target_unique[c]).sum()

    is_monotonic = False
    while not is_monotonic:

        # 卡方 逻辑
        while True:  # 内循环

            ###bmt
            # 判断卡方分箱是否同时满足最小分箱占比和箱子个数
            if len(grouped) <= n_bins and t_min_np(t_cols_sum_axis_1_np(grouped)) >= min_sample_rate:
                break
            ###bmt

            # 计算每一组的卡方
            l = len(grouped) - 1
            chi_list = np.zeros(l, dtype=float)
            chi_min = np.inf
            # chi_ix = []
            for i in range(l):
                chi = 0
                couple = grouped[i:i + 2, :]
                total = t_sum_np(couple)
                cols = t_cols_sum_axis_0_np(couple)
                # t_cols_sum_axis_1(couple)
                # rows = x
                rows = t_cols_sum_axis_1_np(couple)

                for j in range(couple.shape[0]):
                    for k in range(couple.shape[1]):
                        e = rows[j] * cols[k] / total
                        if e != 0:
                            chi += (couple[j, k] - e) ** 2 / e

                # 平衡卡方值
                if balance:
                    chi *= total

                chi_list[i] = chi

                if chi == chi_min:
                    chi_ix.append(i)
                    continue

                if chi < chi_min:
                    chi_min = chi
                    chi_ix = [i]

                # if chi < chi_min:
                #     chi_min = chi

            # 当最小值大于阈值时中断循环
            if min_threshold and chi_min > min_threshold:
                break

            # 获取最小卡方值的那组索引
            min_ix = np.array(chi_ix)
            # min_ix = np.where(chi_list == chi_min)[0]

            # 获取需要删除的索引
            drop_ix = min_ix + 1

            # 按索引合并
            retain_ix = min_ix[0]
            last_ix = retain_ix
            for ix in min_ix:
                # set a new group
                if ix - last_ix > 1:
                    retain_ix = ix

                # 将所有连续索引合并为一组
                for p in range(grouped.shape[1]):
                    grouped[retain_ix, p] = grouped[retain_ix, p] + grouped[ix + 1, p]

                last_ix = ix

            # 删除分组
            grouped = np.delete(grouped, drop_ix, axis=0)
            feature_unique = np.delete(feature_unique, drop_ix)

        # 卡方 逻辑

        splits = feature_unique[1:]

        is_monotonic = True
        if is_need_monotonic:
            x_bins = split_points_to_bin(feature, splits)  # 返回0、1、2、3这样的数值分箱
            bin_badrate_li, _ = bin_badrate(x_bins, target)

            # 不需要判断，内循环里面已经判断了
            # is_monotonic = True if _ >= min_sample_rate else False  # 判断最小分箱占比是否满足

            is_monotonic = pd.Series(bin_badrate_li).is_monotonic_decreasing or pd.Series(
                bin_badrate_li).is_monotonic_increasing

        n_bins = len(splits) + 1  # 初始n_bins为10，对应的splits为9。需要+1，在后面的n_bins -= 1后，n_bins才会是9
        if n_bins <= 2:
            break
        n_bins -= 1

    if is_empty_bin and empty_mask.any():
        splits = np.append(splits, np.nan)

    return splits


@support_dataframe(require_target=False)
def bin_method_run(feature, target=None, method='dt', return_bin=False, **kwargs):
    """
    对数据进行分箱
    Args:
        feature (array-like) : 某个x特征
        target (array-like) :  目标值y变量
        method (str): 分箱方法；'dt'、'chi'、'equal_freq'、'kmeans'四种供选择
        return_bin (bool): 是否返回分箱后的数据
        min_sample_rate (number) : 每个箱子的最小占比
        n_bins (int): 需要分成几箱
        is_need_monotonic (bool) : 是否强制单调
        is_empty_bin (bool) : 是否将空箱单独归为一个箱子
        min_threshold (number): 最小的卡方阀值

    Returns:
        array: 分割点
        array: 原始数据用分箱点替换后的数据

    """

    if method == 'dt':
        splits = dt_bin(feature, target, **kwargs)
    elif method == 'chi':
        splits = chi_bin(feature, target, **kwargs)
    elif method == 'equal_freq':
        splits = equal_freq_bin(feature, target, **kwargs)
    elif method == 'kmeans':
        splits = kmeans_bin(feature, target, **kwargs)
    else:
        splits = np.array([])

    ##返回splits
    if return_bin:
        bins = np.zeros(len(feature))
        if len(splits):
            if np.isnan(splits[-1]):
                mask = pd.isna(feature)
                bins[~mask] = split_points_to_bin(feature[~mask], splits[:-1])
                bins[mask] = len(splits)
            else:
                bins = split_points_to_bin(feature, splits)

        # return splits, pd.Series(bins, name=feature.name)
        return splits, bins

    return splits


def best_binning(df, x_list=[], target='target', **kwargs):
    """
    最优分箱
    Args:
        df (DataFrame) : 需要分箱的数据集
        x_list (list): 需要分箱的特征列表
        target (str): 目标变量
        **kwargs: 'dt'、'chi'、'equal_freq'、'kmeans'四种分箱方法的分箱参数

    Returns:

    """
    from .transformer import FeatureBin

    assert df[target].isin([0, 1]).all(), 'ERROR: :-) {} :-) 目标变量不是0/1值，请检查！！！'.format(target)
    iv_inflection_arr = []
    cutoff_dic = {}
    # for method in ['equal_freq', 'chi', 'dt', 'kmeans']:
    for method in ['equal_freq', 'chi', 'dt']:
        log.info("正在执行最优分箱之 [{}] ".format(method))

        fb = FeatureBin()
        fb.fit(df[x_list], df[target], method=method, **kwargs)
        cutoff_dic[method] = fb.splits_dict
        var_iv_inflection_df = calc_iv_and_inflection_point(fb.transform(df, labels=True)[x_list + [target]],
                                                            target=target, bin_func=method)
        iv_inflection_arr.append(var_iv_inflection_df.reset_index())

        log.info("执行最优分箱之 [{}] over!!!".format(method))

    # 分析获取最优分箱的结果
    iv_inflection_df = pd.concat(iv_inflection_arr, axis=0)
    best_binning_result = iv_inflection_df.groupby('var_name').apply(
        lambda x: x.sort_values(['inflection_point_num', 'IV', 'bin_count'], ascending=[True, False, False]).head(
            1)).set_index("var_name")

    # 找到各个变量最优分箱的分箱方法
    best_binning_func_index = best_binning_result['bin_func'].to_dict()
    best_cutoff = {k: cutoff_dic[v][k] for k, v in best_binning_func_index.items()}
    fb.manual_bin(best_cutoff)

    # 最优分箱，多返回一个数据集记录选择最优的过程
    best_binning_result = best_binning_result.reset_index('var_name')
    return fb, best_binning_result



#!/usr/bin/env python
# ! -*- coding: utf-8 -*-

'''
@File: autobmt_lr_tutorial_code.py
@Author: RyanZheng
@Email: ryan.zhengrp@gmail.com
@Created Time on: 2022-07-11
'''

import pandas as pd

from autobmt.auto_build_scorecard import AutoBuildScoreCard

##**************************************************虚构现实数据例子**************************************************
##**************************************************虚构现实数据例子**************************************************
##**************************************************虚构现实数据例子**************************************************

##**************************************************自动构建lr模型**************************************************

###TODO 注意修改，读取建模数据
data = pd.read_csv('./example_data/tutorial_data.csv')
data_dict = pd.read_excel('./example_data/tutorial_data数据字典.xlsx')  # 读取数据字典，非必要
###TODO 注意修改，读取建模数据

###TODO 注意修改
client_batch = 'TT01p1'
key, target, data_type = 'APP_ID_C', 'target', 'type'  # key是主键字段名，target是目标变量y的字段名，data_type是train、test数据集标识的字段名
ml_res_save_path = './example_model_result/{}'.format(client_batch)  # 模型结果保存的位置
###TODO 注意修改

###TODO 下面代码基本可以不用动
# 初始化
autobtmodel = AutoBuildScoreCard(datasets=data,  # 训练模型的数据集
                                 fea_names=list(data.columns),  # 数据集的字段名
                                 target=target,  # 目标变量y字段名
                                 key=key,  # 主键字段名
                                 data_type=data_type,  # train、test数据集标识的字段名
                                 no_feature_names=[key, target, data_type] + ['apply_time'],
                                 # 数据集中不用于开发模型的特征字段名，即除了x特征的其它字段名
                                 ml_res_save_path=ml_res_save_path,  # 建模相关结果保存路径
                                 data_dict=data_dict,  # 数据字典，非必要，有则添加，无则不要此参数
                                 AB={'A': 404.65547021957406, 'B': 72.13475204444818},    #自定义的大A，大B；非必要，有默认的
                                 positive_corr=False,   #数与模型预测的概率值是否正相关。默认False，负相关，即概率约高，分数越低
                                 )

# ###训练模型
model, in_model_fea = autobtmodel.fit(empty_threhold=0.95,  # 特征的缺失值大于该阀值的特征会被剔除
                                      iv_threhold=0.02,  # iv小于该阀值的特征会被剔除
                                      corr_threhold=0.7,  # 相关性大于等于该阀值的特征会被剔除
                                      psi_threhold=0.05  # psi大于等于该阀值的特征会被剔除
                                      )
###TODO 上面代码基本可以不用动


# ##**************************************************读取未来需要预测的数据，使用训练好的模型进行预测**************************************************
#
# ###TODO 注意修改，读取未来需要预测的数据
# data = pd.read_csv(
#     './example_data/tutorial_data.csv')
# ###TODO 注意修改，读取未来需要预测的数据
#
# # 未来有新数据过来，使用训练好的模型进行预测
# offline_pred_res = AutoBuildScoreCard.predict(to_pred_df=data,  # 未来需要预测的数据，id+特征即可
#                            model_path='./example_model_result/TT01p1/20231116115935_35',
#                            # 训练好的模型路径
#                            )
# offline_pred_res.to_csv(
#     './example_model_result/TT01p1/20231116115935_35/offline_pred_res.csv',
#     index=False)
#
# ##**************************************************虚构现实数据例子**************************************************
# ##**************************************************虚构现实数据例子**************************************************
# ##**************************************************虚构现实数据例子**************************************************



#!/usr/bin/env python

"""The setup script."""

import re
from setuptools import setup, find_packages

NAME = 'autobmt'

with open('README.md') as readme_file:
    readme = readme_file.read()


def get_version():
    with open(f"{NAME}/__init__.py") as f:
        return re.search(r'\d+\.\d+\.\d+', f.read()).group()


def get_requirements(stage=None):
    file_name = 'requirements'

    if stage is not None:
        file_name = f"{file_name}-{stage}"

    requirements = []
    with open(f"{file_name}.txt", 'r') as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith('-'):
                continue

            requirements.append(line)

    return requirements


test_requirements = ['pytest', ]

setup(
    author="RyanZheng",
    author_email='zhengruiping000@163.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="a modeling tool that automatically builds scorecards and tree models.",
    install_requires=get_requirements(),
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords='autobmt',
    name=NAME,
    packages=find_packages(include=['autobmt', 'autobmt.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ZhengRyan/autobmt',
    version=get_version(),
    zip_safe=False,
)
